@@ -17,7 +17,7 @@ use Crypt::CBC;
 require Exporter;
 use vars qw($VERSION $DEBUG @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
 @ISA = qw(Exporter);
-$VERSION = '0.11';
+$VERSION = '0.12';
 
 BEGIN {
   $DEBUG='no';
@@ -388,7 +388,7 @@ sub Server {
   ErrWarn(\@warning, "NO CRL Check implemented");
   ErrWarn(\@warning, "NO Certificate Purpose Check implemented");
   ErrWarn(\@warning, "NO CA signing policy check implemented");
-  for (my $i=0;$i < $#Errors;$i++ ) { if ( @{ @Errors[$i] } ) { ErrWarn(\@error,"Error verifying client's certificate chain (depth $i): ".join(", ",@{ @Errors[$i] })); } }
+  for (my $i=0;$i < $#Errors;$i++ ) { if ( @{ $Errors[$i] } ) { ErrWarn(\@error,"Error verifying client's certificate chain (depth $i): ".join(", ",@{ $Errors[$i] })); } }
   ErrWarn(\@error,"Client's Certificate is not trusted by server") unless ($Trust[-1]);
   return { Errors => \@error, Warnings => \@warning } if (@error);
 
@@ -655,7 +655,7 @@ sub Get {
   ErrWarn(\@warning, "NO CRL Check implemented");
   ErrWarn(\@warning, "NO Certificate Purpose Check implemented");
   ErrWarn(\@warning, "NO CA signing policy check implemented");
-  for (my $i=0;$i < $#Errors;$i++ ) { if ( @{ @Errors[$i] } ) { ErrWarn(\@error,"Error verifying server's certificate chain (depth $i): ".join(", ",@{ @Errors[$i] })); } }
+  for (my $i=0;$i < $#Errors;$i++ ) { if ( @{ $Errors[$i] } ) { ErrWarn(\@error,"Error verifying server's certificate chain (depth $i): ".join(", ",@{ $Errors[$i] })); } }
   ErrWarn(\@error,"Server's Certificate is not trusted by client") unless ($Trust[-1]);
   return { Errors => \@error, Warnings => \@warning } if (@error);
 
@@ -848,6 +848,340 @@ sub Get {
 }####Endof sub Get
 
 
+sub List {
+  my %context=%{ $_[0]};
+  my @error; my @warning;
+  if ( $] < 5.004 ) { ErrWarn(\@warning, "Perl version is old; random seed is not good"); }
+
+  if ( ! defined $context{'Server'} )   { ErrWarn(\@error, "Server not Specified"); }
+  if ( ! defined $context{'Port'} )     { ErrWarn(\@error, "Port not Specified"); }
+  if ( ! defined $context{'VO'} )       { ErrWarn(\@error, "VO not Specified"); }
+
+# IO::SOCKET::SSL may optionally use a cert and key on the file system, we too;
+  if ( ! defined $context{'Cert'} && ! defined $context{'CertFile'} ) { ErrWarn(\@error, "Certificate not Specified"); }
+  if ( ! defined $context{'Key'} && ! defined $context{'KeyFile'} )   { ErrWarn(\@error, "Key not Specified"); }
+  if ( ! defined $context{'Cert'} && defined $context{'CertFile'} && ! -r $context{'CertFile'} ) { ErrWarn(\@error, "Certificate file unreadable"); }
+  if ( ! defined $context{'Key'} && defined $context{'KeyFile'} && ! -r $context{'KeyFile'} ) { ErrWarn(\@error, "Key file unreadable"); }
+
+  if ( @error > 0 ) { return { Errors => \@error, Warnings => \@warning }; }
+
+  my $Server       = (($context{'Server'}   =~ /^([a-z0-9_.-]+)$/) ? $& : undef);
+  my $VO           = (($context{'VO'}       =~ /^([a-z0-9_.-]+)$/) ? $& : undef);
+  my $Port         = (($context{'Port'}     =~ /^([0-9]{1,5})$/ && $context{'Port'} < 65536) ? $& : 7512);
+
+# Barf if data is not good
+  if ( ! defined $Server )         { ErrWarn(\@error, "Bad VOMS server string"); }
+  if ( ! defined $VO )             { ErrWarn(\@error, "Bad VO string"); }
+  if ( ! defined $Port )           { ErrWarn(\@error, "Bad Port"); }
+
+  my @certs; my $key;
+  if ( ref($context{'Cert'}) eq "ARRAY" ) { @certs = @{ $context{'Cert'} }; }
+  elsif ( defined($context{'Cert'}) and ref($context{'Cert'}) eq "" )   { @certs = ( $context{'Cert'} ); } #might consider a function to seperate concatenated DERs 
+  elsif ( defined($context{'Cert'}) )     { ErrWarn(\@error, "Certs Argument was not a reference to an array nor a scalar"); }
+  else { @certs = ( readCert($context{'CertFile'}) );}
+  if ( ! @certs ) { ErrWarn(\@error, "Unable to get any user certs."); }
+  foreach my $i (0 .. $#certs) { if ( $certs[$i] !~ /^\x30/s ) { ErrWarn(\@error, "Supplied certificate (\@context{'Cert'}[$i]) $certs[$i] not in DER format"); } }
+
+  if ( defined $context{'Key'} && $context{'Key'} !~ /^\x30/s ) { ErrWarn(\@error, "Supplied Key not in DER format"); }
+  if ( defined $context{'Key'} ) { $key=$context{'Key'}; }
+  else { $key=readPrivateKey($context{'KeyFile'}); }
+  if (! defined($key) ) { ErrWarn(\@error, "Unable to get user key."); }
+
+  my @CAdirs;
+  if ( defined $context{'CAdirs'} ) {
+    if ( ref($context{'CAdirs'}) eq "ARRAY" ) { @CAdirs = @{ $context{'CAdirs'} }; }
+    elsif ( ref($context{'Cert'}) eq "" )   { @CAdirs = split(':',$context{'CAdirs'}); }
+    else { ErrWarn(\@error, "CAdirs Argument was not a reference to an array nor a scalar"); }
+  }
+
+  foreach my $i (0 .. $#CAdirs) { if ( ! -d $CAdirs[$i] ) { ErrWarn(\@error, "Supplied CA directory (\@context{'CAdirs'}[$i]) is not a directory"); } }
+
+  if ( @error > 0 ) { return { Errors => \@error, Warnings => \@warning }; }
+
+#Need CAdirs
+#==========================================================
+  if ( ! @CAdirs and $ENV{X509_CERT_DIR} ) {
+    if ( -d $ENV{X509_CERT_DIR} and $ENV{X509_CERT_DIR} =~ /^(.*)$/) { push @CAdirs,$1; }
+    else { return { Errors => ErrWarn('X509_CERT_DIR defined but it is not a directory'), Warnings => \@warning }; }
+  }
+  elsif ( ! @CAdirs ) {
+    if ( -d $CAdir ) { push @CAdirs, $CAdir; ErrWarn(\@warning,"no CAdir specified Using $CAdir"); }
+    else { return {   Errors => ErrWarn('No CAdir found'), Warnings => \@warning }; }
+  }
+
+# Get details from Cert and Key
+  my %certinfo = %{ VOMS::Lite::X509::Examine( $certs[0], { SubjectDN=>"", IssuerDN=>"" }) };
+  my %chain    = %{ VOMS::Lite::CertKeyHelper::buildchain( { trustedCAdirs => \@CAdirs, suppliedcerts => \@certs } ) };
+  my @chain    = @{ $chain{'Certs'} };
+  my $UserDN   = $chain{'EndEntityDN'};
+  my $UserIDN  = $chain{'EndEntityIssuerDN'};
+  my %keyinfo  = %{ VOMS::Lite::KEY::Examine( $key, { Keymodulus=>"",KeyprivateExponent=>"" }) };
+  my $Keymod   = Hex($keyinfo{'Keymodulus'});
+  my $Keyexp   = Hex($keyinfo{'KeyprivateExponent'});
+  my $DN       = $certinfo{'SubjectDN'};
+  my $IDN      = $certinfo{'IssuerDN'};
+
+#Open a socket to the server
+  my $sock = new IO::Socket::INET( PeerAddr => $Server, PeerPort => $Port, Proto => 'tcp', Type => SOCK_STREAM); 
+  if ( ! defined ($sock) ) { return { Errors => ErrWarn("Unable to establish a connection to $Server:$Port"), Warnings => \@warning }; }
+  $sock->autoflush(1);
+
+#######################################
+#Construct Initial Components required for SSL 
+#Random
+  my $hextime               = DecToHex(time);
+  my $time                  = Bin($hextime);
+  my $rnd = "XXXXXXXXXXXXXXXXXXXXXXXXXXXX"; $rnd =~ s/./chr(int(rand 256))/ge;#not so good on Win32 32000 cycle reported
+  my $hexrnd                = Hex($rnd);
+  my $hexrandom             = $hextime.$hexrnd;
+  my $hexsession            = "00";   #none - new session
+  my $hexcypher_vec         = "000a";  #Ciphersuite TLS_RSA_WITH_3DES_EDE_CBC_SHA
+  my $hexcypher_suits       = sprintf("%04s",DecToHex(length($hexcypher_vec)/2)).$hexcypher_vec;
+  my $hexcompression_vec    = "00"; #CompressionMeth  - Use none no need data minimal + this is a lite implementation
+  my $hexcompression        = sprintf("%02s",DecToHex(length($hexcompression_vec)/2)).$hexcompression_vec;
+  my $hexssl_version        = "0300";
+  my $hexhandshake_parts    = $hexssl_version.$hexrandom.$hexsession.$hexcypher_suits.$hexcompression;
+  my $hexhello              = handShake("01",$hexhandshake_parts);
+  my $clienthello           = Bin($hexhello); 
+  my $hexhellorecord        = recordLayer("16",$hexhello);
+
+#######################################
+#Send Client Hello Record
+  my $hellorecord           = Bin($hexhellorecord); #Client Hello Record ready to send
+  Write($sock,$hellorecord);
+
+#######################################
+# Listen for ServerHello
+  my $response=Read($sock);
+  my $hexserverhellorecord=Hex($response);
+  my $records="";
+  while (length($response) > 0) {
+    my $lenstr=substr($response,0,5,'');
+    my $len;
+    if ($lenstr =~ /^\x16\x03\x00(.)(.)/s ) { $len=ord($1)*256+ord($2); }
+    else { return { Errors => ErrWarn("Malformed SSL header from server while waiting for SSL ServerHello messages"), Warnings => \@warning }; }
+    $records.=substr($response,0,$len,'');
+  }
+  my $serverhello=$records;
+
+#######################################
+# Decode ServerHandshake Messages
+  my %Hand = ParseRecords($records);
+
+# Get the host certificate - 1, Verify VOMS cert 2, key material for pre-master key exchange later
+  my $certcont = $Hand{11};
+  my $lcerts = substr($certcont,0,3,'');
+  my @HOSTcerts=();
+  while ( length($certcont)>3 && ($lcert=substr($certcont,0,3,''))) { 
+    if ($lcert =~ /(.)(.)(.)/s ) { push @HOSTcerts, substr($certcont,0,(ord($1)*65536)+(ord($2)*256)+ord($3),''); }
+  }
+  return { Errors => ErrWarn("No cert returned from server"), Warnings => \@warning } unless (@HOSTcerts);
+
+# Build VOMS Server Chain Locally and check it's trusted
+  my %Chain  = %{ buildchain( { trustedCAdirs => \@CAdirs, suppliedcerts => \@HOSTcerts } ) };
+  my @Trust  = @{ $Chain{TrustedCA} };
+  my @Errors = @{ $Chain{Errors} };
+  my $EEC           =    $Chain{EndEntityCert};
+  ErrWarn(\@warning, "NO CRL Check implemented");
+  ErrWarn(\@warning, "NO Certificate Purpose Check implemented");
+  ErrWarn(\@warning, "NO CA signing policy check implemented");
+  for (my $i=0;$i < $#Errors;$i++ ) { if ( @{ $Errors[$i] } ) { ErrWarn(\@error,"Error verifying server's certificate chain (depth $i): ".join(", ",@{ $Errors[$i] })); } }
+  ErrWarn(\@error,"Server's Certificate is not trusted by client") unless ($Trust[-1]);
+  return { Errors => \@error, Warnings => \@warning } if (@error);
+
+# Extract Server Cert and Key details
+  my %ServerCertInfo= %{ VOMS::Lite::X509::Examine( $EEC, { SubjectDN=>"", IssuerDN=>"", Keymodulus=>"", KeypublicExponent=>"", subjectAltName=>"" }) };
+  my $ServerDN=$ServerCertInfo{'SubjectDN'};
+  my @SubjecyAltNames=@{ $ServerCertInfo{'subjectAltNameArray'} };
+  debug("Server DN",$ServerDN);
+# Check server certificate matches ! Not matching against wildcards 
+  if ($ServerDN !~ m#/CN=($Server)(/|$)#) {
+    my $match=0;
+# Else Check Subject Alt Name matches
+    foreach (@SubjecyAltNames) { debug("AltName",$_); if ($_ eq "dNSName=$Server") {$match=1; ErrWarn(\@warning,"Using SubjectAltName to Match VOMS Server Certificate")} }
+    return { Errors => ErrWarn("Server Distinguished name mismatch expecting Certificate name containing CN=$Server got $ServerDN"), Warnings => \@warning } 
+     if (!$match && $Server ne "localhost");
+  }
+
+# Get ServerHello bits and pieces
+  my $sHello                  = $Hand{2};
+  my $sHelloVer               = substr($sHello,0,2,'');
+  my $sHelloTime              = substr($sHello,0,4,'');
+  my $sHelloRand              = substr($sHello,0,28,'');
+  my $sHelloIDlen             = ord(substr($sHello,0,1,''));
+  my $sHelloSessionID         = substr($sHello,0,$sHelloIDlen,'');
+  my $sHelloCypherSuite       = substr($sHello,0,2,'');
+  my $sHelloCompressionMethod = substr($sHello,0,1,'');
+  debug("Session ID",$sHelloSessionID);
+
+# Get Certificate Request (request for authN)
+  my $sRequest=$Hand{13};
+  my $ReqTypesCount=ord(substr($sRequest,0,1,''));
+  my $RSAOK=0;
+  my @ReqTypes; for (my $a=0; $a<$ReqTypesCount; $a++) { push @ReqTypes,substr($sRequest,0,1,''); if ($ReqTypes[-1] eq "\x01") {$RSAOK=1;} }
+  my $ReqDNNamesLen=(ord(substr($sRequest,0,1,''))*256)+ord(substr($sRequest,0,1,''));
+  my $ReqDistinguishedNames=substr($sRequest,0,$ReqDNNamesLen,'');
+  my @ReqASN1DN;
+  while ( length($ReqDistinguishedNames) > 0 ) {
+    my $DNLen=(ord(substr($ReqDistinguishedNames,0,1,''))*256)+ord(substr($ReqDistinguishedNames,0,1,''));
+    push @ReqASN1DN,substr($ReqDistinguishedNames,0,$DNLen,'');
+  }
+
+# Get ServerHelloDone
+  my $sHelloDone=$Hand{14};
+
+# Check for RSA
+  if ($RSAOK==0) { return { Errors => ErrWarn("Server does not support RSA AuthN"), Warnings => \@warning }; } 
+
+# 1, Check acceptable DNs for Certificates
+  my $GotCA=0;
+  foreach (@ReqASN1DN) {
+    my $X509subject=$_;
+    my @ASN1SubjectDNIndex=ASN1Index($X509subject);
+    shift @ASN1SubjectDNIndex;
+    my $SubjectDN="";
+    while (@ASN1SubjectDNIndex) {
+      my ($CLASS,$CONSTRUCTED,$TAG,$HEADSTART,$HEADLEN,$CHUNKLEN)=(0,0,0,0,0);
+      until ($TAG == 6 ) { ($CLASS,$CONSTRUCTED,$TAG,$HEADSTART,$HEADLEN,$CHUNKLEN) = @{shift @ASN1SubjectDNIndex}; }
+      my $OID=substr($X509subject,($HEADSTART+$HEADLEN),$CHUNKLEN);
+      ($CLASS,$CONSTRUCTED,$TAG,$HEADSTART,$HEADLEN,$CHUNKLEN) = @{shift @ASN1SubjectDNIndex};
+      my $Value=substr($X509subject,($HEADSTART+$HEADLEN),$CHUNKLEN);
+      $SubjectDN.="/".VOMS::Lite::CertKeyHelper::OIDtoDNattrib(ASN1OIDtoOID($OID))."=$Value";
+    }
+    if ($UserIDN eq $SubjectDN) { debug("MATCHED CA",$SubjectDN); $GotCA=1; } 
+    else                        { debug("        CA:", $SubjectDN); }
+  }
+  if ( @ReqASN1DN == 0 ) { ErrWarn(\@warning, "VOMS server does not tell me what CAs are supported"); }
+  elsif ( $GotCA==0 )    { return { Errors => ErrWarn("VOMS server does not support your CA"), Warnings => \@warning }; }
+
+#########################################################
+#Talk to the server again
+
+##Construct Certificate Message Record
+  my $hexcertarray =""; 
+  foreach my $cert (@chain) { $hexcertarray.=sprintf("%06s",DecToHex(length($cert))).Hex($cert);}
+  my $hexcertobj      = sprintf("%06s",DecToHex(length($hexcertarray)/2)).$hexcertarray;
+  my $hexcertmesg     = handShake("0b",$hexcertobj);
+  my $hexcertsmessagerecord=recordLayer("16",$hexcertmesg);
+  my $certmesg = pack("H*", $hexcertmesg); 
+
+##ClientKeyExchange
+#PreMasterSecret
+  my $prernd='X' x 46;  $prernd =~ s/./chr(int(rand 256))/ge;
+  my $hexpremastersecret='0300'.Hex($prernd);
+
+#Derive Key Material
+  my %KEYS=Keygen($hexpremastersecret,$hexrandom,Hex($sHelloTime.$sHelloRand));
+
+#Make Client Key Exchange
+  my $hexEncPMSecret   = rsaencrypt($hexpremastersecret,Hex($ServerCertInfo{'KeypublicExponent'}),Hex($ServerCertInfo{'Keymodulus'}));
+  my $hexClientKeyExchange  = handShake("10",$hexEncPMSecret);    #### USE ME FOR CertificateVarify
+  my $hexClientKeyExchangeMessageRecord = recordLayer("16",$hexClientKeyExchange);
+  my $ClientKeyExchangeMessage = Bin($hexClientKeyExchange); #### ClientCertificate for handshakemessages needs to be without record layer
+
+#CertificateVerify
+  my $Hmsgs     = $clienthello.$serverhello.$certmesg.$ClientKeyExchangeMessage;
+  my $verifymac = MAC($KEYS{'MS'},$Hmsgs);
+
+  my $hexsignedcertificateverify        = rsasign($verifymac,$Keyexp,$Keymod);
+  my $hexwrappedsignedcertificateverify = sprintf("%04s",DecToHex(length($hexsignedcertificateverify)/2)).$hexsignedcertificateverify;
+  my $hexcertificateverify              = handShake('0f',$hexwrappedsignedcertificateverify);
+  my $certificateverify                 = Bin($hexcertificateverify);
+  my $hex_ssl_certificateverifyrecord   = recordLayer("16",$hexcertificateverify);
+
+##########################################################################
+## Switch to Encrypted Session -- change_cipher_spec message
+#Select algorythm for key exchange -- Must be RSA
+  my $hexkeyselection="140300000101";
+
+# Switch to tripple des (the only one I support) and send finished message
+
+  $Hmsgs.=$certificateverify;
+  my $unechexfinished=handShake("14",MAC($KEYS{'MS'},$Hmsgs."CLNT"));
+
+#Set up the Client and Server DES3-EDE sessions
+  my $Ccipher = Crypt::CBC->new( -literal_key => 1, -padding => 'null', -key => $KEYS{'CK'},  -iv => $KEYS{'CIV'}, -header => 'none', -cipher => 'DES_EDE3' );
+  my $Scipher = Crypt::CBC->new( -literal_key => 1, -padding => 'null', -key => $KEYS{'SK'},  -iv => $KEYS{'SIV'}, -header => 'none', -cipher => 'DES_EDE3' );
+  my $Cseq=0;
+  my $Sseq=0;
+
+  my $hex_finishedrecord = &Encrypt(Bin(recordLayer("16",$unechexfinished)),$KEYS{'CMAC'},$pad1sha,$pad2sha,$Cseq,$Ccipher);
+  my $hex_finished       = $hex_finishedrecord;
+  $hex_finished         =~ s/^..........//s;
+  my $finished           = Bin($hex_finished);
+
+#############################################################################
+#Send Records
+  Write($sock,Bin($hexcertsmessagerecord.$hexClientKeyExchangeMessageRecord.$hex_ssl_certificateverifyrecord.$hexkeyselection.$hex_finishedrecord));
+
+#############################################################################
+#Receive Response
+  $response=Read($sock);
+  if ( $response =~ /^\x14\x03\x00\x00\x01\x01/ ) {  $response =~ s/......//; } #Change Cypher Spec -- what we expect
+  else { return { Errors => ErrWarn("Error: Expecting SSL Change Cypher Spec Message, got something else."), Warnings => \@warning }; }
+  my $serverfinishedhex=Hex(&Decrypt($response,$KEYS{'SMAC'},$pad1sha,$pad2sha,$Sseq,$Scipher));
+
+# Check response from server is a valid finished message
+  $Hmsgs.=Bin($unechexfinished);
+  my $unencryptedfinishedmsg = handShake("14",MAC($KEYS{'MS'},$Hmsgs."SRVR"));
+  if ($unencryptedfinishedmsg ne $serverfinishedhex) { return { Errors => ErrWarn("Failed to decrypt Server Finished Message"), Warnings => \@warning }; }
+
+######################### 
+# Send no delegation byte
+  my $msg='0';
+  my $emesg        = &Encrypt(Bin(recordLayer("17",Hex($msg))),$KEYS{'CMAC'},$pad1sha,$pad2sha,$Cseq,$Ccipher);
+  $senddata        = Bin($emesg);
+  Write($sock,$senddata);
+
+############## 
+# Send request
+#<xml version="1.0" encoding = "US-ASCII"?>
+#  <voms>
+#    <command>COMMAND</command>+
+#    <order>ORDER</order>?
+#    <targets>TARGETS</targets>?
+#    <lifetime>N</lifetime>?
+#    <base64>B</base64>?
+#    <version>V</version>?
+#  </voms>
+# COMMAND:  G/vo.name(/subgroup)* {All relevant} | Rrolename {All relevant} | B/vo.name(/subgroup)*:rolename | A {All} | M {List} | /vo.name(/subgroup)*/Role=rolename
+#  my $msg='<?xml version="1.0" encoding = "US-ASCII"?><voms><command>G/ngs.ac.uk/ops</command><base64>1</base64><version>4</version><lifetime>43200</lifetime></voms>';
+#  my $msg='<?xml version="1.0" encoding = "US-ASCII"?><voms><command>ROperations</command><base64>1</base64><version>4</version><lifetime>43200</lifetime></voms>';
+
+  my $cmds = "<command>N</command>";
+
+#  $msg='<?xml version="1.0" encoding = "US-ASCII"?><voms>'.$cmds.'<base64>1</base64><version>4</version><lifetime>'.$lifetime.'</lifetime></voms>';
+  $msg='<?xml version="1.0" encoding = "US-ASCII"?><voms>'.$cmds.'</voms>';
+  $emesg         = &Encrypt(Bin(recordLayer("17",Hex($msg))),$KEYS{'CMAC'},$pad1sha,$pad2sha,$Cseq,$Ccipher);
+  my $senddata2     = Bin($emesg);
+  Write($sock,$senddata2);
+
+###############
+# Read Response
+  $response=Read($sock);
+  if ( $response !~ /^\x17/ ) { return { Errors => ErrWarn("Expecting Encrypted Application Data, got something else"), Warnings => \@warning }; }
+  my $apdata=&Decrypt($response,$KEYS{'SMAC'},$pad1sha,$pad2sha,$Sseq,$Scipher);
+  my ($bitstr) = $apdata =~ /^.*<bitstr>([^<]*)<\/bitstr>.*$/;
+  $bitstr =~ s/[^a-zA-Z0-9_\+=\/\[\]-]//g;
+  $apdata =~ s|<error><item><number>[0-9]{4,}</number><message>([^<]*)</message></item></error>|ErrWarn(\@error,"Error from VOMS Server: \"$1\"")|ge;
+  $apdata =~ s|<error><item><number>[0-9]{1,3}</number><message>([^<]*)</message></item></error>|ErrWarn(\@warning,"Warning from VOMS Server: \"$1\"")|ge;
+  if ( @error > 0 ) { return { Errors => \@error, Warnings => \@warning }; }
+
+  my $details=VOMS::Lite::Base64::Decode($bitstr,"VOMS");
+  close($sock);
+
+  my @FQANs   = grep { /^\// } split(/\n/s,$details);
+  my @details = grep { /^\/$VO/ } @FQANs;
+  my @others  = grep { ! /^\/$VO/ } @FQANs;
+  if (@others) { my ($badvo) = $others[0] =~ m|^/([^/]+)|; push @warning => "The server seems to support a different VO: $badvo"; }
+  for (my $i=0;$i<@details;$i++) {
+    $details[$i] =~ s/\/(?:Capability=NULL|Role=NULL)$//;
+    $details[$i] =~ s/\/(?:Capability=NULL|Role=NULL)$//;
+  }
+
+  return { FQANs=>\@details, Warnings => \@warning };
+}####Endof sub List
+
 ######################################################################
 sub Encrypt { #this routine modifies $_[4,5] Encrypt($rec,$MACSecret,$pad1sha,$pad2sha,$seq,$cipher)
   my ($rec,$MACSecret,$pad1sha,$pad2sha)=@_;
@@ -973,6 +1307,7 @@ None.
 =head1 Also See
 
 https://twiki.cnaf.infn.it/cgi-bin/twiki/view/VOMS/VOMSProtocol
+(NB Command "M" should read "N")
 http://glite.cvs.cern.ch/cgi-bin/glite.cgi/org.glite.security.voms
 
 RFC3281 and the VOMS Attribute Specification document from the OGSA Athuz Workin
